# -*- coding: utf-8 -*-
# <nbformat>3.0</nbformat>

# <headingcell level=1>

# probfit Basic Tutorial

# <markdowncell>

# [probfit](http://iminuit.github.io/probfit/) is a modeling / fitting package to be used together with [iminuit](http://iminuit.github.com/iminuit/).
# 
# This tutorial is a fast-paced introduction to the probfit features:
# 
# * built-in common models: polynomial, gaussian, ...
# * build-in common fit statistics: chi^2, binned and unbinned likelihood
# * tools to get your fits to converge and check the results: try_uml, draw, draw_residuals, ...
# * tools to help you implement your own models and fit statistics: Normalize, Extended, integrate_1d, ...
# 
# Please start this notebook with the ``ipython --pylab=inline`` option to get inline plots.

# <codecell>

# We assume you have executed this cell in all the following examples
import numpy as np
import matplotlib.pyplot as plt
import iminuit
import probfit

# <markdowncell>

# In your own code you can explicitly import what you need to save
# typing in interactive sessions, e.g.
# 
#     from iminuit import Minuit, describe
#     from probfit import gaussian, BinnedLH
# 
# We don't do this here, we only import `iminuit` and `probfit` into our
# namespace so that it is clear to you which functions and classes come
# from which package while reading the code below.

# <markdowncell>

# ## Chi^2 straight line fit
# 
# We can't really call this a fitting package without being able to fit a straight line, right?

# <codecell>

# Let's make a straight line with gaussian(mu=0, sigma=1) noise
np.random.seed(0)
x = np.linspace(0, 10, 20) 
y = 3 * x + 15 + np.random.randn(len(x))
err = np.ones(len(x))
plt.errorbar(x, y, err, fmt='.');

# <codecell>

# Let's define our line.
# First argument has to be the independent variable,
# arguments after that are shape parameters.
def line(x, m, c): # define it to be parabolic or whatever you like
    return m * x + c

# <codecell>

iminuit.describe(line)

# <codecell>

# Define a chi^2 cost function
chi2 = probfit.Chi2Regression(line, x, y, err)

# <codecell>

# Chi2Regression is just a callable object; nothing special about it
iminuit.describe(chi2)

# <codecell>

# minimize it
# yes, it gives you a heads up that you didn't give it initial value
# we can ignore it for now
minuit = iminuit.Minuit(chi2) # see iminuit tutorial on how to give initial value/range/error
minuit.migrad(); # MIGRAD is a very stable robust minimization method
# you can look at your terminal to see what it is doing;

# <codecell>

# The output above is a pretty-printed summary of the fit results from
# minuit.print_fmin()
# which was automatically called by iminuit.Minuit.migrad() after running MIGRAD.

# Let's see our results as Python dictionaries ...
print(minuit.values)
print(minuit.errors)

# <markdowncell>

# #### Parabolic error
# is calculated using the second derivative at the minimum
# This is good in most cases where the uncertainty is symmetric not much correlation
# exists. Migrad usually got this accurately but if you want ot be sure
# call `minuit.hesse()` after calling `minuit.migrad()`.
# 
# #### Minos Error
# is obtained by scanning the chi^2 or likelihood profile and find the point
# where chi^2 is increased by `minuit.errordef`. Note that in the Minuit documentation
# and output `errordef` is often called `up` ... it's the same thing.
# 
# #### What `errordef` should I use?
# 
# As explained in the Minuit documentation you should use:
# 
# * `errordef = 1` for chi^2 fits
# * `errordef = 0.5` for likelihood fits
# 
# `errordef=1` is the default, so you only have to set it to `errordef=0.5`
# if you are defining a likelihood cost function (if you don't your HESSE and MINOS errors will be incorrect).
# `probfit` helps you by defining a `default_errordef()` attribute on the
# cost function classes, which is automatically detected by the `Minuit` constructor
# and can be used to set `Minuit.errordef` correctly, so that users can't forget.
# Classes used in this tutorial:
# 
# * `probfit.Chi2Regression.get_errordef()` and `probfit.BinnedChi2.get_errordef()` return 1.
# * `probfit.BinnedLH.get_errordef()` and `probfit.UnbinnedLH.get_errordef()` return 0.5.

# <codecell>

# Let's visualize our line
chi2.draw(minuit)
# looks good;

# <codecell>

# Sometimes we want the error matrix (a.k.a. covariance matrix)
print('error matrix:')
print(minuit.matrix())
# or the correlation matrix
print('correlation matrix:')
print(minuit.matrix(correlation=True))
# or a pretty html representation
# Note that `print_matrix()` shows the correlation matrix, not the error matrix
minuit.print_matrix()

# <markdowncell>

# ## Binned Poisson likelihood fit of a Gaussian distribution
# In high energy physics, we usually want to fit a distribution to a histogram. Let's look at simple Gaussian distribution.

# <codecell>

# First let's make some example data
np.random.seed(0)
data = np.random.randn(10000) * 4 + 1
# sigma = 4 and mean = 1
plt.hist(data, bins=100, histtype='step');

# <codecell>

# Define your PDF / model
def gauss_pdf(x, mu, sigma):
    """Normalized Gaussian"""
    return 1 / np.sqrt(2 * np.pi) / sigma * np.exp(-(x - mu) ** 2 / 2. / sigma ** 2)

# <codecell>

# Build your cost function
# Here we use binned likelihood
binned_likelihood = probfit.BinnedLH(gauss_pdf, data)

# <codecell>

# Create the minuit
# and give an initial value for the sigma parameter
minuit = iminuit.Minuit(binned_likelihood, sigma=3)
# Remember: minuit.errordef is automatically set to 0.5
# as required for likelihood fits (this was explained above)
binned_likelihood.draw(minuit);

# <codecell>

minuit.migrad()
# Like in all binned fit with long zero tail. It will have to do something about the zero bin
# probfit.BinnedLH does handle them gracefully but will give you a warning;

# <codecell>

# Visually check if the fit succeeded by plotting the model over the data
binned_likelihood.draw(minuit) # uncertainty is given by symmetric Poisson;

# <codecell>

# Let's see the result
print('Value: {}'.format(minuit.values))
print('Error: {}'.format(minuit.errors))

# <codecell>

# That printout can get out of hand quickly
minuit.print_fmin()
# Also print the correlation matrix
minuit.print_matrix()

# <codecell>

# Looking at a likelihood profile is a good method
# to check that the reported errors make sense
minuit.draw_mnprofile('mu');

# <codecell>

# Plot a 2d contour error
# You can notice that it takes some time to draw
# We will this is because our PDF is defined in Python
# We will show how to speed this up later
minuit.draw_mncontour('mu', 'sigma');

# <markdowncell>

# ## Chi^2 fit of a Gaussian distribution
# 
# Let's explore another popular cost function chi^2.
# Chi^2 is bad when you have bin with 0.
# ROOT just ignore.
# ROOFIT does something I don't remember.
# But it's best to avoid using chi^2 when you have bin with 0 count.

# <codecell>

# We will use the same data as in the previous example
np.random.seed(0)
data = np.random.randn(10000) * 4 + 1
# sigma = 4 and mean = 1
plt.hist(data, bins=100, histtype='step');

# <codecell>

# We will use the same PDF as in the previous example
def gauss_pdf(x, mu, sigma):
    """Normalized Gaussian"""
    return 1 / np.sqrt(2 * np.pi) / sigma * np.exp(-(x - mu) **2 / 2. / sigma ** 2)

# <codecell>

# Binned chi^2 fit only makes sense (for now) for extended PDFs
# probfit.Extended adds a norm parameter with name 'N'
extended_gauss_pdf = probfit.Extended(gauss_pdf)

# <codecell>

# Describe the function signature
iminuit.describe(extended_gauss_pdf)

# <codecell>

# Chi^2 distribution fit is really bad for distribution with long tail
# since when bin count=0... poisson error=0 and blows up chi^2
# so give it some range
chi2 = probfit.BinnedChi2(extended_gauss_pdf, data, bound=(-7,10))
# This time we use the pedantic=False option to tell Minuit
# that we don't want warnings about parameters without initial
# value or step size.
# And print_level=0 means that no output is generated
minuit = iminuit.Minuit(chi2, sigma=1, pedantic=False, print_level=0)
minuit.migrad();

# <codecell>

# Now let's look at the results
minuit.print_fmin()
minuit.print_matrix()
chi2.draw(minuit);

# <markdowncell>

# ## Fast unbinned likelihood fit Cython
# 
# Unbinned likelihood is computationally very very expensive if you have a lot of data.
# It's now a good time that we talk about how to speed things up with [Cython](http://cython.org).

# <codecell>

# We will use the same data as in the previous example
np.random.seed(0)
data = np.random.randn(10000) * 4 + 1
# sigma = 4 and mean = 1
plt.hist(data, bins=100, histtype='step');

# <codecell>

# We want to speed things up with Cython
%load_ext cythonmagic

# <codecell>

%%cython
# Same gaussian distribution but now written in Cython
# The %%cython IPython does the following:
# * Call Cython to generate C code for a Python C extension.
# * Compile it into a Python C extension (a shared library)
# * Load it into the current namespace
# If you don't understand these things, don't worry, it basically means:
# * Get full-metal speed easily
cimport cython
from libc.math cimport exp, M_PI, sqrt
@cython.binding(True) # IMPORTANT: this tells Cython to dump the function signature
def gauss_pdf_cython(double x, double mu, double sigma):
    return 1 / sqrt(2 * M_PI) / sigma * exp(-(x - mu) ** 2 / 2. / sigma ** 2)

# <codecell>

# Define the unbinned likelihood cost function 
unbinned_likelihood = probfit.UnbinnedLH(gauss_pdf_cython, data)

# <codecell>

minuit = iminuit.Minuit(unbinned_likelihood, sigma=2, pedantic=False, print_level=0)
# Remember: minuit.errordef is automatically set to 0.5
# as required for likelihood fits (this was explained above)
minuit.migrad() # yes: amazingly fast
unbinned_likelihood.show(minuit)
minuit.print_fmin()
minuit.print_matrix() 

# <codecell>

# Remember how slow draw_mnprofile() was in the last example?
# Now it's super fast (even though the unbinned
# likelihood computation is more compute-intensive).
minuit.draw_mnprofile('mu');

# <markdowncell>

# But you really don't have to write your own gaussian, there are tons of builtin functions written in Cython for you.

# <codecell>

# Here's how you can list them
import probfit.pdf
print(dir(probfit.pdf))
print(iminuit.describe(probfit.pdf.gaussian))
print(type(probfit.pdf.gaussian))
# But actually they are always all imported into the main probfit
# namespace, so we'll keep using the simpler probfit.gaussian instead of
# probfit.pdf.gaussian here.

# <codecell>

unbinned_likelihood = probfit.UnbinnedLH(probfit.gaussian, data)
minuit = iminuit.Minuit(unbinned_likelihood, sigma=2, pedantic=False)
# Remember: minuit.errordef is automatically set to 0.5
# as required for likelihood fits (this was explained above)
minuit.migrad() # yes: amazingly fast
unbinned_likelihood.draw(minuit, show_errbars='normal') # control how fit is displayed too;

# <codecell>

# Draw the difference between data and PDF
plt.figure(figsize=(13,4))
plt.subplot(121)
unbinned_likelihood.draw_residual(minuit)
plt.subplot(122)
unbinned_likelihood.draw_residual(minuit, show_errbars=True, errbar_algo='sumw2', norm=True)

# <markdowncell>

# ##But... We can't normalize everything analytically and how to generate toy sample from PDF
# 
# When fitting distribution to a PDF, one of the common problem that we run into is normalization.
# Not all function is analytically integrable on the range of our interest.
# 
# Let's look at an example: the [Crystal Ball function](http://en.wikipedia.org/wiki/Crystal_Ball_function).
# It's simply a gaussian with a power law tail ... normally found in energy deposited in crystals ...
# impossible to normalize analytically and normalization will depend on shape parameters.

# <codecell>

numpy.random.seed(0)
bound = (-1, 2)
data = probfit.gen_toy(probfit.crystalball, 10000, bound=bound, alpha=1., n=2., mean=1., sigma=0.3, quiet=False)
# quiet=False tells gen_toy to plot out original function
# toy histogram and poisson error from both orignal distribution and toy

# <codecell>

# To fit this function as a distribution we need to normalize
# so that is becomes a PDF ober the range we consider here.
# We do this with the probfit.Normalized functor, which implements
# the trapezoid numerical integration method with a simple cache mechanism
normalized_crystalball = probfit.Normalized(probfit.crystalball, bound)
# this can also bedone with decorator
# @probfit.normalized(bound)
# def my_function(x, blah):
#    return something
pars = 1.0, 1, 2, 1, 0.3
print('function: {}'.format(probfit.crystalball(*pars)))
print('     pdf: {}'.format(normalized_crystalball(*pars)))

# <codecell>

# The normalized version has the same signature as the non-normalized version
print(iminuit.describe(probfit.crystalball))
print(iminuit.describe(normalized_crystalball))

# <codecell>

# We can fit the normalized function in the usual way ...
unbinned_likelihood = probfit.UnbinnedLH(normalized_crystalball, data)
start_pars = dict(alpha=1, n=2.1, mean=1.2, sigma=0.3)
minuit = iminuit.Minuit(unbinned_likelihood, **start_pars)
# Remember: minuit.errordef is automatically set to 0.5
# as required for likelihood fits (this was explained above)
minuit.migrad() # yes: amazingly fast Normalize is written in Cython
unbinned_likelihood.show(minuit)
# The Crystal Ball function is notorious for its sensitivity on the 'n' parameter
# probfit give you a heads up where it might have float overflow;

# <markdowncell>

# ## But what if I know the analytical integral formula for my distribution?
# 
# `probfit` checks for a method called `integrate` with the signature `integrate(bound, nint, *arg)` to
# compute definite integrals for given `bound` and `nint` (pieces of integral this is normally ignored)
# and the rest will be passed as positional argument.
# 
# For some `probfit` built-in distributions analytical formulae have been implemented.

# <codecell>

def line(x, m, c):
    return m * x + c

# compute integral of line from x=(0,1) using 10 intevals with m=1. and c=2.
# all probfit internal use this
# no integrate method available probfit use simpson3/8
print(probfit.integrate1d(line, (0, 1), 10, (1., 2.)))

# Let us illustrate the point by forcing it to have integral that's off by
# factor of two
def wrong_line_integrate(bound, nint, m, c):
    a, b = bound
    # I know this is wrong:
    return 2 * (m * (b ** 2 / 2. - a ** 2 / 2.) + c * (b - a))

line.integrate = wrong_line_integrate
# line.integrate = lambda bound, nint, m, c: blah blah # this works too
print(probfit.integrate1d(line, (0, 1), 10, (1., 2.)))

# <headingcell level=2>

# What if things go wrong?

# <markdowncell>

# In this section we show you what happens when your distribution doesn't fit and how you can make it.
# 
# We again use the Crystal Ball distribution as an example, which is notoriously sensitive to initial parameter values.

# <codecell>

unbinned_likelihood = probfit.UnbinnedLH(normalized_crystalball, data)
# No initial values given -> all parameters have default initial value 0
minuit = iminuit.Minuit(unbinned_likelihood)
# Remember: minuit.errordef is automatically set to 0.5
# as required for likelihood fits (this was explained above)
minuit.migrad() # yes: amazingly fast but tons of output on the console
# Remember there is a heads up;

# <codecell>

# This shows that we failed.
# The parameters are still at the default initial values
unbinned_likelihood.show(minuit);

# <codecell>

# These two status flags tell you if the best-fit parameter values
# and the covariance matrix (the parameter errors) are OK.
print(minuit.migrad_ok())
print(minuit.matrix_accurate())

# <markdowncell>

# To make MIGRAD converge we need start parameter values that are roughly correct. Remember that above the same fit converged when we used ::
# 
#     start_pars = dict(alpha=1, n=2.1, mean=1.2, sigma=0.3)
#     minuit = iminuit.Minuit(unbinned_likelihood, **start_pars)
#     
# #### But how can we guess these initial values?
# 
# This is a hard question that doesn't have one simple answer. Visualizing your data and model helps.

# <codecell>

# Try one set of parameters
best_try = probfit.try_uml(normalized_crystalball, data, alpha=1., n=2.1, mean=1.2, sigma=0.3)
print(best_try)

# <codecell>

# Or try multiple sets of parameters
# (too many will just confuse you)
best_try = probfit.try_uml(normalized_crystalball, data, alpha=1., n=2.1, mean=[1.2, 1.1], sigma=[0.3, 0.5])
# try_uml computes the unbinned likelihood for each set of parameters and returns the best
# one as a dictionary.
# This is actually a poor-man's optimization algorithm in itself called grid search
# which is popular to find good start values for other, faster optimization methods like MIGRAD.
print(best_try)

# <headingcell level=2>

# Extended fit: two Gaussians with polynomial background

# <markdowncell>

# Here we show how to create and fit a model that is the sum of several other models.

# <codecell>

# Generate some example data
np.random.seed(0)
data_peak1 = np.random.randn(3000) * 0.2 + 2
data_peak2 = np.random.randn(5000) * 0.1 + 4
data_range = (-2, 5)
data_bg = probfit.gen_toy(lambda x : 4 + 4 * x + x ** 2, 20000, data_range)
data_all = np.concatenate([data_peak1, data_peak2, data_bg])
plt.hist((data_peak1, data_peak2, data_bg, data_all),
         label=['Signal 1', 'Signal 2', 'Background', 'Total'],
         bins=200, histtype='step', range=data_range)
plt.legend(loc='upper left');

# <codecell>

# Using a polynomial to fit a distribution is problematic, because the
# polynomial can assume negative values, which results in NaN (not a number)
# values in the likelihood function.
# To avoid this problem we restrict the fit to the range (0, 5) where
# the polynomial is clearly positive.
fit_range = (0, 5)
normalized_poly = probfit.Normalized(probfit.Polynomial(2), fit_range)
normalized_poly = probfit.Extended(normalized_poly, extname='NBkg')

gauss1 = probfit.Extended(probfit.rename(probfit.gaussian, ['x', 'mu1', 'sigma1']), extname='N1')
gauss2 = probfit.Extended(probfit.rename(probfit.gaussian, ['x', 'mu2', 'sigma2']), extname='N2')

# Define an extended PDF consisting of three components
pdf = probfit.AddPdf(normalized_poly, gauss1, gauss2)

print('normalized_poly: {}'.format(probfit.describe(normalized_poly)))
print('gauss1:          {}'.format(probfit.describe(gauss1)))
print('gauss2:          {}'.format(probfit.describe(gauss2)))
print('pdf:             {}'.format(probfit.describe(pdf)))

# <codecell>

# Define the cost function in the usual way ...
binned_likelihood = probfit.BinnedLH(pdf, data_all, bins=200, extended=True, bound=fit_range)

# This is a quite complex fit (11 free parameters!), so we need good starting values.
# Actually we even need to set an initial parameter error
# for 'mu1' and 'mu2' to make MIGRAD converge.
# The initial parameter error is used as the initial step size in the minimization.
pars = dict(mu1=1.9, error_mu1=0.1, sigma1=0.2, N1=3000,
            mu2=4.1, error_mu2=0.1, sigma2=0.1, N2=5000,
            c_0=4, c_1=4, c_2=1, NBkg=20000)
minuit = iminuit.Minuit(binned_likelihood, pedantic=False, print_level=0, **pars)
# You can see that the model already roughly matches the data
binned_likelihood.draw(minuit, parts=True);

# <codecell>

# This can take a while ... the likelihood is evaluated a few 100 times
# (and each time the distributions are evaluated, including the
# numerical computation of the normalizing integrals)
minuit.migrad();

# <codecell>

binned_likelihood.show(minuit, parts=True);
minuit.print_fmin()
minuit.print_matrix()

# <markdowncell>

# Note the red upper left corner in the correlation matrix above?
# 
# It shows that the three polynomial parameters `c_0`, `c_1` and `c_2` are highly correlated?
# The reason is that we put a constraint on the polynomial to be normalized over the fit range:
# 
#     fit_range = (0, 5)
#     normalized_poly = probfit.Normalized(probfit.Polynomial(2), fit_range)
#     normalized_poly = probfit.Extended(normalized_poly, extname='NBkg')
# 
# To resolve this problem you could simply use a non-normalized and non-extended polynomial to model the background. We won't do this here, though ...

# <markdowncell>

# ## Custom Drawing
# 
# The `draw()` and `show()` method we provide is intended to just give you a quick look at your fit.
# 
# To make a custom drawing you can use the return value of `draw()` and `show()`.

# <codecell>

# You should copy & paste the return tuple from the `draw` docstring ...
((data_edges, datay), (errorp, errorm), (total_pdf_x, total_pdf_y), parts) = binned_likelihood.draw(minuit, parts=True);
# ... now we have everything to make our own plot

# <codecell>

# Now make the plot as pretty as you like, e.g. with matplotlib.
plt.figure(figsize=(8, 5))
plt.errorbar(probfit.mid(data_edges), datay, errorp, fmt='.', capsize=0, color='Gray', label='Data')
plt.plot(total_pdf_x, total_pdf_y, color='blue', lw=2, label='Total Model')
colors = ['orange', 'purple', 'DarkGreen']
labels = ['Background', 'Signal 1', 'Signal 2']
for color, label, part in zip(colors, labels, parts):
    x, y = part
    plt.plot(x, y, ls='--', color=color, label=label)
plt.grid(True)
plt.legend(loc='upper left');

# <markdowncell>

# ## Simultaneous fit to several data sets
# 
# Sometimes, what we want to fit is the sum of likelihood /chi^2 of two PDFs for two different datasets that share some parameters.
# 
# In this example, we will fit two Gaussian distributions where we know that the widths are the same
# but the peaks are at different places.

# <codecell>

# Generate some example data
np.random.seed(0)
data1 = np.random.randn(10000) + 3 # mean =  3, sigma = 1
data2 = np.random.randn(10000) - 2 # mean = -2, sigma = 1
plt.figure(figsize=(12,4))
plt.subplot(121)
plt.hist(data1, bins=100, range=(-7, 7), histtype='step', label='data1')
plt.legend()
plt.subplot(122)
plt.hist(data2, bins=100, range=(-7, 7), histtype='step', label='data2')
plt.legend();

# <codecell>

# There is nothing special about built-in cost function
# except some utility function like draw and show
likelihood1 = probfit.UnbinnedLH(probfit.rename(probfit.gaussian, ('x', 'mean2', 'sigma')), data1)
likelihood2 = probfit.UnbinnedLH(probfit.gaussian, data2)
simultaneous_likelihood = probfit.SimultaneousFit(likelihood1, likelihood2)
print(probfit.describe(likelihood1))
print(probfit.describe(likelihood2))
# Note that the simultaneous likelihood has only 3 parameters, because the
# 'sigma' parameter is tied (i.e. linked to always be the same).
print(probfit.describe(simultaneous_likelihood))

# <codecell>

# Ah, the beauty of Minuit ... it doesn't care what your cost funtion is ...
# you can use it to fit (i.e. compute optimal parameters and parameter errors) anything.
minuit = iminuit.Minuit(simultaneous_likelihood, sigma=0.5, pedantic=False, print_level=0)
# Well, there's one thing we have to tell Minuit so that it can compute parameter errors,
# and that is the value of `errordef`, a.k.a. `up` (explained above).
# This is a likelihood fit, so we need `errordef = 0.5` and not the default `errordef = 1`:
minuit.errordef = 0.5

# <codecell>

# Run the fit and print the results
minuit.migrad();
minuit.print_fmin()
minuit.print_matrix()

# <codecell>

simultaneous_likelihood.draw(minuit);

# <markdowncell>

# ## Blinding parameters
# 
# Often, an analyst would like to avoid looking at the result of the fitted parameter(s) before he/she finalized the analysis in order to avoid biases due to the prejudice of the analyst. Probfit provids a transformation function that hides the true value(s) of the parameter(s). The transformation function requires a string to set the seed of the random number generator, and a scale to smear the parameter(s) using a Gaussian.

# <codecell>

from probfit import UnbinnedLH, BlindFunc, rename, AddPdfNorm
from probfit import gaussian
from iminuit import Minuit, describe
from probfit import gen_toy

# <codecell>

g0= rename(gaussian, ['x', 'm0', 's0'])
g1= rename(gaussian, ['x', 'm1', 's1'])
pdf= AddPdfNorm(g0,g1)
describe(pdf)

# <codecell>

seed(0)
toydata = gen_toy(pdf, 1000,(-10,10), m0=-2, m1=2, s0=1, s1=1, f_0=0.3, quiet=False)

# <codecell>

inipars= dict(m0=0, m1=0, s0=1, s1=1, f_0=0.5, error_m0=0.1, error_m1=0.1, error_s0=0.1, error_s1=0.1, error_f_0=0.1)

# <codecell>

# Normal fit
uh1= UnbinnedLH(pdf, toydata)
m1= Minuit(uh1, print_level=1, **inipars)
m1.migrad();
uh1.draw();
print m1.values

# <codecell>

# Blind one parameter
uh2= UnbinnedLH( BlindFunc(pdf, toblind='m1', seedstring='some_random_stuff', width=0.5, signflip=False), toydata)
m2= Minuit(uh2, print_level=1, **inipars)
m2.migrad();
uh2.draw();
print m2.values

# <codecell>

# Blind more than one parameter. They will be shifted by the same amount
uh3= UnbinnedLH( BlindFunc(pdf, ['m0','m1'], seedstring='some_random_stuff', width=0.5, signflip=False), toydata)
m3= Minuit(uh3, print_level=1, **inipars)
m3.migrad();
uh3.draw();
print m3.values

# <codecell>

print m1.values
print m2.values
print m3.values
print 
print m1.errors
print m2.errors
print m3.errors

# <codecell>

print m3.values['m0']-m1.values['m0']
print m3.values['m1']-m1.values['m1']

# <codecell>

# Now it's your turn ...
# try and apply probfit / iminuit and to your modeling / fitting task! 




class ProbfitWarning(RuntimeWarning):
    pass


class SmallIntegralWarning(ProbfitWarning):
    pass


class SmallDivisionWarning(ProbfitWarning):
    pass


class SharedExtendeeExtenderParameter(ProbfitWarning):
    pass


class LogWarning(ProbfitWarning):
    pass



import numpy as np
from ._libstat import _vector_apply
from .py23_compat import range


def fwhm_f(f, range, arg=None, bins=1000):
    arg = tuple() if arg is None else arg

    x = np.linspace(range[0], range[1], bins)
    y = _vector_apply(f, x, arg)
    imax = np.argmax(y)
    ymax = y[imax]
    rs = y[imax:] - ymax / 2.0
    ls = y[:imax] - ymax / 2.0
    xl = 0
    xr = 0

    il = first_neg(ls, 'l')
    # print il,x[il],ls[il],x[il+1],ls[il+1]
    xl = xintercept(x[il], ls[il], x[il + 1], ls[il + 1])

    ir = first_neg(rs, 'r')
    # print ir,x[imax+ir],rs[ir],x[ir+1],rs[ir+1]
    xr = xintercept(x[imax + ir], rs[ir], x[imax + ir - 1], rs[ir - 1])

    return (xl, xr)


def xintercept_tuple(t0, t1):
    return xintercept(t0[0], t0[1], t1[0], t1[1])


def xintercept(x0, y0, x1, y1):
    m = (y1 - y0) / (x1 - x0)
    return -y0 / m + x0


def first_neg(y, direction='r'):
    if direction == 'l':
        xlist = range(len(y) - 1, -1, -1)
    else:
        xlist = range(len(y))
    ret = 0
    found = False
    for i in xlist:
        if y[i] < 0:
            ret = i
            found = True
            break
    if not found:
        raise ValueError('They are all positive what are you tying to find?')
    return ret



import numpy as np
from ._libstat import _vector_apply


def mid(x):
    return (x[:-1] + x[1:]) / 2


def minmax(x):
    return min(x), max(x)


# for converting numpy array to double
def float2double(a):
    if a is None or a.dtype == np.float64:
        return a
    else:
        return a.astype(np.float64)


def vector_apply(f, x, *arg):
    """
    Apply **f** to array **x** with given arguments fast.

    This is a fast version of::

        np.array([f(xi,*arg) for xi in x ])

    useful when you try to plot something.
    """
    return _vector_apply(f, x, arg)



# -*- coding: utf-8 -*-
# Plotting is on python since this will make it much easier to debug and adjsut
# no need to recompile everytime i change graph color....
# needs a serious refactor
from math import sqrt, ceil, floor
from warnings import warn
import numpy as np
from matplotlib import pyplot as plt
from .nputil import mid, minmax, vector_apply
from .util import parse_arg, describe


def draw_simultaneous(self, minuit=None, args=None, errors=None, **kwds):
    numf = len(self.allf)
    ret = []
    numraw = sqrt(numf)
    numcol = ceil(numraw)
    numrow = floor(numraw) if floor(numraw) * numcol >= numf else ceil(numraw)

    for i in range(numf):
        plt.subplot(numrow, numcol, i + 1)
        part_args, part_errors = self.args_and_error_for(i, minuit, args, errors)
        ret.append(self.allf[i].draw(args=part_args, errors=part_errors, **kwds))

    return ret


def _get_args_and_errors(self, minuit=None, args=None, errors=None):
    """
    consistent algorithm to get argument and errors
    1) get it from minuit if minuit is available
    2) if not get it from args and errors
    2.1) if args is dict parse it.
    3) if all else fail get it from self.last_arg
    """
    ret_arg = None
    ret_error = None
    if minuit is not None:  # case 1
        ret_arg = minuit.args
        ret_error = minuit.errors
        return ret_arg, ret_error

    # no minuit specified use args and errors
    if args is not None:
        if isinstance(args, dict):
            ret_arg = parse_arg(self, args)
        else:
            ret_arg = args
    else:  # case 3
        ret_arg = self.last_arg

    if errors is not None:
        ret_error = errors

    return ret_arg, ret_error


def _param_text(parameters, arg, error):
    txt = u''
    for i, (k, v) in enumerate(zip(parameters, arg)):
        txt += u'%s = %5.4g' % (k, v)
        if error is not None:
            txt += u'±%5.4g' % error[k]
        txt += u'\n'
    return txt


    # from UML


def draw_ulh(self, minuit=None, bins=100, ax=None, bound=None,
             parmloc=(0.05, 0.95), nfbins=200, print_par=True, grid=True,
             args=None, errors=None, parts=False, show_errbars='normal',
             no_plot=False):
    data_ret = None
    error_ret = None
    total_ret = None
    part_ret = []

    ax = plt.gca() if ax is None and not no_plot else ax

    arg, error = _get_args_and_errors(self, minuit, args, errors)

    n, e = np.histogram(self.data, bins=bins, range=bound, weights=self.weights)
    dataint = (n * np.diff(e)).sum()
    data_ret = (e, n)

    if not show_errbars:
        if not no_plot:
            pp = ax.hist(mid(e), bins=e, weights=n, histtype='step')
        error_ret = (np.sqrt(n), np.sqrt(n))
    else:
        w2 = None
        if show_errbars == 'normal':
            w2 = n
            error_ret = (np.sqrt(n), np.sqrt(n))
        elif show_errbars == 'sumw2':
            weights = None
            if self.weights != None:
                weights = self.weights ** 2
            w2, e = np.histogram(self.data, bins=e, weights=weights)
            error_ret = (np.sqrt(w2), np.sqrt(w2))
        else:
            raise ValueError('show_errbars must be \'normal\' or \'sumw2\'')
        if not no_plot:
            ax.errorbar(mid(e), n, np.sqrt(w2), fmt='b.', capsize=0,
                        zorder=0)

    # bound = (e[0], e[-1])
    draw_arg = [('lw', 2), ('zorder', 2)]
    if not parts:
        draw_arg.append(('color', 'r'))

    # Draw pdf with finer bins
    ef = np.linspace(e[0], e[-1], nfbins + 1)
    scale = dataint if not self.extended else nfbins / float(bins)
    total_ret = draw_pdf_with_edges(self.f, arg, ef, ax=ax, density=not self.extended, scale=scale,
                                    **dict(draw_arg))

    if parts:
        f_parts = getattr(self.f, 'parts', None)
        if f_parts is not None:
            for p in f_parts():
                ret = draw_pdf_with_edges(p, arg, ef, ax=ax, scale=scale, density=not self.extended, no_plot=no_plot)
                part_ret.append(ret)
    if not no_plot:
        ax.grid(grid)

    txt = _param_text(describe(self), arg, error)
    if not no_plot and print_par:
        ax.text(parmloc[0], parmloc[1], txt, ha='left', va='top',
                transform=ax.transAxes)
    return (data_ret, error_ret, total_ret, part_ret)


def draw_residual(x, y, yerr, xerr,
                  show_errbars=True, ax=None,
                  zero_line=True, grid=True,
                  **kwargs):
    """Draw a residual plot on the axis.

    By default, if show_errbars if True, residuals are drawn as blue points
    with errorbars with no endcaps. If show_errbars is False, residuals are
    drawn as a bar graph with black bars.

    **Arguments**

        - **x** array of numbers, x-coordinates

        - **y** array of numbers, y-coordinates

        - **yerr** array of numbers, the uncertainty on the y-values

        - **xerr** array of numbers, the uncertainty on the x-values

        - **show_errbars** If True, draw the data as a bar plot, else as an
          errorbar plot

        - **ax** Optional matplotlib axis instance on which to draw the plot

        - **zero_line** If True, draw a red line at :math:`y = 0` along the
          full extent in :math:`x`

        - **grid** If True, draw gridlines

        - **kwargs** passed to ``ax.errorbar`` (if ``show_errbars`` is True) or
          ``ax.bar`` (if ``show_errbars`` if False)

    **Returns**

    The matplotlib axis instance the plot was drawn on.
    """
    ax = plt.gca() if ax is None else ax

    if show_errbars:
        plotopts = dict(fmt='b.', capsize=0)
        plotopts.update(kwargs)
        pp = ax.errorbar(x, y, yerr, xerr, zorder=0, **plotopts)
    else:
        plotopts = dict(color='k')
        plotopts.update(kwargs)
        pp = ax.bar(x - xerr, y, width=2*xerr, **plotopts)

    if zero_line:
        ax.plot([x[0] - xerr[0], x[-1] + xerr[-1]], [0, 0], 'r-', zorder=2)

    # Take the `grid` kwarg to mean 'add a grid if True'; if grid is False and
    # we called ax.grid(False) then any existing grid on ax would be turned off
    if grid:
        ax.grid(grid)

    return ax


def draw_residual_ulh(self, minuit=None, bins=100, ax=None, bound=None,
                      parmloc=(0.05, 0.95), print_par=False,
                      args=None, errors=None, errbar_algo='normal', norm=False,
                      **kwargs):
    ax = plt.gca() if ax is None else ax

    arg, error = _get_args_and_errors(self, minuit, args, errors)

    n, e = np.histogram(self.data, bins=bins, range=bound, weights=self.weights)
    dataint = (n * np.diff(e)).sum()
    scale = dataint if not self.extended else 1.0
    w2 = None
    if errbar_algo == 'normal':
        w2 = n
    elif errbar_algo == 'sumw2':
        weights = None
        if self.weights != None:
            weights = self.weights ** 2
        w2, e = np.histogram(self.data, bins=e, weights=weights)
    else:
        raise ValueError('errbar_algo must be \'normal\' or \'sumw2\'')
    yerr = np.sqrt(w2)

    arg = parse_arg(self.f, arg, 1) if isinstance(arg, dict) else arg
    yf = vector_apply(self.f, mid(e), *arg)
    yf *= (scale * np.diff(e) if self.extended else scale)
    n = n - yf
    if norm:
        sel = yerr > 0
        n[sel] /= yerr[sel]
        yerr = np.ones(len(yerr))

    ax = draw_residual(mid(e), n, yerr, np.diff(e)/2.0, ax=ax, **kwargs)

    txt = _param_text(describe(self), arg, error)
    if print_par:
        ax.text(parmloc[0], parmloc[1], txt, ha='left', va='top',
                transform=ax.transAxes)


        # from chi2 regression


def draw_x2(self, minuit=None, ax=None, parmloc=(0.05, 0.95), print_par=True,
            args=None, errors=None, grid=True, parts=False, no_plot=False):
    data_ret = None
    error_ret = None
    total_ret = None
    part_ret = []

    ax = plt.gca() if ax is None and not no_plot else ax

    arg, error = _get_args_and_errors(self, minuit, args, errors)

    x = self.x
    y = self.y
    data_err = self.error

    data_ret = x, y
    if data_err is None:
        if not no_plot: ax.plot(x, y, '+')
        err_ret = (np.ones(len(self.x)), np.ones(len(self.x)))
    else:
        if not no_plot: ax.errorbar(x, y, data_err, fmt='.', zorder=0)
        err_ret = (data_err, data_err)
    draw_arg = [('lw', 2), ('zorder', 2)]
    draw_arg.append(('color', 'r'))

    total_ret = draw_pdf_with_midpoints(self.f, arg, x, ax=ax, no_plot=no_plot, **dict(draw_arg))

    if not no_plot:
        ax.grid(grid)

    txt = _param_text(describe(self), arg, error)

    chi2 = self(*arg)
    if self.ndof > 0:
        txt += u'chi2/ndof = %5.4g(%5.4g/%d)' % (chi2 / self.ndof, chi2, self.ndof)
    else:
        txt += u'chi2/ndof = (%5.4g/%d)' % (chi2, self.ndof)

    if parts:
        f_parts = getattr(self.f, 'parts', None)
        if f_parts is not None:
            for p in f_parts():
                tmp = draw_pdf_with_midpoints(p, arg, x, ax=ax, no_plot=no_plot, **dict(draw_arg))
                part_ret.append(tmp)

    if print_par and not no_plot:
        ax.text(parmloc[0], parmloc[1], txt, ha='left', va='top',
                transform=ax.transAxes)

    return (data_ret, error_ret, total_ret, part_ret)


def draw_x2_residual(self, minuit=None, ax=None, args=None, errors=None, grid=True,
                     norm=False):
    ax = plt.gca() if ax is None else ax

    arg, error = _get_args_and_errors(self, minuit, args, errors)

    x = self.x
    y = self.y
    data_err = self.error
    f = self.f

    arg = parse_arg(f, arg, 1) if isinstance(arg, dict) else arg
    yf = vector_apply(f, x, *arg)

    yplot = y - yf
    eplot = data_err if data_err is not None else np.zeros(len(x))
    if norm:
        if data_err is None:
            warn(RuntimeWarning('No error on data points; cannot normalize to error'))
        else:
            yplot = yplot / data_err
            eplot = data_err / data_err
    ax.errorbar(x, yplot, eplot, fmt='b+')
    ax.grid(grid)


    # from binned chi2


def draw_bx2(self, minuit=None, parmloc=(0.05, 0.95), nfbins=500, ax=None,
             print_par=True, args=None, errors=None, parts=False, grid=True,
             no_plot=False):
    data_ret = None
    error_ret = None
    total_ret = None
    part_ret = []

    ax = plt.gca() if ax is None and not no_plot else ax

    arg, error = _get_args_and_errors(self, minuit, args, errors)

    m = mid(self.edges)

    if not no_plot:
        ax.errorbar(m, self.h, self.err, fmt='.', zorder=0)
    data_ret = (self.edges, self.h)
    error_ret = (self.err, self.err)

    bound = (self.edges[0], self.edges[-1])

    scale = nfbins / float(self.bins)  # scale back to bins

    draw_arg = [('lw', 2), ('zorder', 2)]

    if not parts:
        draw_arg.append(('color', 'r'))

    total_ret = draw_pdf(self.f, arg, bins=nfbins, bound=bound, ax=ax, density=False,
                         scale=scale, no_plot=no_plot, **dict(draw_arg))

    if parts:
        f_parts = getattr(self.f, 'parts', None)
        if f_parts is not None:
            for p in f_parts():
                tmp = draw_pdf(p, arg, bound=bound, bins=nfbins, ax=ax, density=False,
                               scale=scale, no_plot=no_plot)
                part_ret.append(tmp)

    if not no_plot:
        ax.grid(grid)

    txt = _param_text(describe(self), arg, error)

    chi2 = self(*arg)
    if self.ndof > 0:
        txt += u'chi2/ndof = %5.4g(%5.4g/%d)' % (chi2 / self.ndof, chi2, self.ndof)
    else:
        txt += u'chi2/ndof = (%5.4g/%d)' % (chi2, self.ndof)

    if print_par and not no_plot:
        ax.text(parmloc[0], parmloc[1], txt, ha='left', va='top',
                transform=ax.transAxes)

    return (data_ret, error_ret, total_ret, part_ret)


    # from binnedLH


def draw_blh(self, minuit=None, parmloc=(0.05, 0.95),
             nfbins=1000, ax=None, print_par=True, grid=True,
             args=None, errors=None, parts=False, no_plot=False):
    data_ret = None
    error_ret = None
    total_ret = None
    part_ret = []

    ax = plt.gca() if ax is None and not no_plot else ax

    arg, error = _get_args_and_errors(self, minuit, args, errors)

    m = mid(self.edges)

    if self.use_w2:
        err = np.sqrt(self.w2)
    else:
        err = np.sqrt(self.h)

    n = np.copy(self.h)
    dataint = (n * np.diff(self.edges)).sum()
    scale = dataint if not self.extended else 1.0

    if not no_plot:
        ax.errorbar(m, n, err, fmt='.', zorder=0)
    data_ret = (self.edges, n)
    error_ret = (err, err)

    draw_arg = [('lw', 2), ('zorder', 2)]
    if not parts:
        draw_arg.append(('color', 'r'))
    bound = (self.edges[0], self.edges[-1])

    # scale back to bins
    if self.extended:
        scale = nfbins / float(self.bins)
    total_ret = draw_pdf(self.f, arg, bins=nfbins, bound=bound, ax=ax, density=not self.extended,
                         scale=scale, no_plot=no_plot, **dict(draw_arg))
    if parts:
        f_parts = getattr(self.f, 'parts', None)
        if f_parts is not None:
            for p in f_parts():
                tmp = draw_pdf(p, arg, bins=nfbins, bound=bound, ax=ax,
                               density=not self.extended, scale=scale, no_plot=no_plot)
                part_ret.append(tmp)
    if not no_plot:
        ax.grid(grid)

    txt = _param_text(describe(self), arg, error)

    if print_par and not no_plot:
        ax.text(parmloc[0], parmloc[1], txt, ha='left', va='top',
                transform=ax.transAxes)

    return (data_ret, error_ret, total_ret, part_ret)


def draw_residual_blh(self, minuit=None, parmloc=(0.05, 0.95),
                      ax=None, print_par=False, args=None, errors=None,
                      norm=False, **kwargs):
    ax = plt.gca() if ax is None else ax

    arg, error = _get_args_and_errors(self, minuit, args, errors)

    m = mid(self.edges)

    if self.use_w2:
        err = np.sqrt(self.w2)
    else:
        err = np.sqrt(self.h)

    n = np.copy(self.h)
    dataint = (n * np.diff(self.edges)).sum()
    scale = dataint if not self.extended else 1.0

    arg = parse_arg(self.f, arg, 1) if isinstance(arg, dict) else arg
    yf = vector_apply(self.f, m, *arg)
    yf *= (scale * np.diff(self.edges) if self.extended else scale)
    n = n - yf
    if norm:
        sel = err > 0
        n[sel] /= err[sel]
        err = np.ones(len(err))

    ax = draw_residual(m, n, err, np.diff(self.edges)/2.0, ax=ax, **kwargs)

    txt = _param_text(describe(self), arg, error)

    if print_par:
        ax.text(parmloc[0], parmloc[1], txt, ha='left', va='top',
                transform=ax.transAxes)


def draw_compare(f, arg, edges, data, errors=None, ax=None, grid=True, normed=False, parts=False):
    """
    TODO: this needs to be rewritten
    """
    # arg is either map or tuple
    ax = plt.gca() if ax is None else ax
    arg = parse_arg(f, arg, 1) if isinstance(arg, dict) else arg
    x = (edges[:-1] + edges[1:]) / 2.0
    bw = np.diff(edges)
    yf = vector_apply(f, x, *arg)
    total = np.sum(data)
    if normed:
        ax.errorbar(x, data / bw / total, errors / bw / total, fmt='.b',
                    zorder=0)
        ax.plot(x, yf, 'r', lw=2, zorder=2)
    else:
        ax.errorbar(x, data, errors, fmt='.b', zorder=0)
        ax.plot(x, yf * bw, 'r', lw=2, zorder=2)

    # now draw the parts
    if parts:
        if not hasattr(f, 'eval_parts'):
            warn(RuntimeWarning('parts is set to True but function does '
                                'not have eval_parts method'))
        else:
            scale = bw if not normed else 1.
            parts_val = list()
            for tx in x:
                val = f.eval_parts(tx, *arg)
                parts_val.append(val)
            py = zip(*parts_val)
            for y in py:
                tmpy = np.array(y)
                ax.plot(x, tmpy * scale, lw=2, alpha=0.5)
    plt.grid(grid)
    return x, yf, data


def draw_normed_pdf(f, arg, bound, bins=100, scale=1.0, density=True, ax=None, **kwds):
    return draw_pdf(f, arg, bound, bins=100, scale=1.0, density=True,
                    normed_pdf=True, ax=ax, **kwds)


def draw_pdf(f, arg, bound, bins=100, scale=1.0, density=True,
             normed_pdf=False, ax=None, **kwds):
    """
    draw pdf with given argument and bounds.

    **Arguments**

        * **f** your pdf. The first argument is assumed to be independent
          variable

        * **arg** argument can be tuple or list

        * **bound** tuple(xmin,xmax)

        * **bins** number of bins to plot pdf. Default 100.

        * **scale** multiply pdf by given number. Default 1.0.

        * **density** plot density instead of expected count in each bin
          (pdf*bin width). Default True.

        * **normed_pdf** Normalize pdf in given bound. Default False

        * The rest of keyword argument will be pass to pyplot.plot

    **Returns**

        x, y of what's being plot
    """
    edges = np.linspace(bound[0], bound[1], bins)
    return draw_pdf_with_edges(f, arg, edges, ax=ax, scale=scale, density=density,
                               normed_pdf=normed_pdf, **kwds)


def draw_pdf_with_edges(f, arg, edges, ax=None, scale=1.0, density=True,
                        normed_pdf=False, **kwds):
    x = (edges[:-1] + edges[1:]) / 2.0
    bw = np.diff(edges)
    scale *= bw if not density else 1.

    return draw_pdf_with_midpoints(f, arg, x, ax=ax, scale=scale,
                                   normed_pdf=normed_pdf, **kwds)


def draw_pdf_with_midpoints(f, arg, x, ax=None, scale=1.0, normed_pdf=False, no_plot=False, **kwds):
    ax = plt.gca() if ax is None and not no_plot else ax
    arg = parse_arg(f, arg, 1) if isinstance(arg, dict) else arg
    yf = vector_apply(f, x, *arg)

    if normed_pdf:
        normed_factor = sum(yf)  # assume equal binwidth
        yf /= normed_factor
    yf *= scale

    if not no_plot:
        ax.plot(x, yf, **kwds)
    return x, yf


    # draw comparison between function given args and data


def draw_compare_hist(f, arg, data, bins=100, bound=None, ax=None, weights=None,
                      normed=False, use_w2=False, parts=False, grid=True):
    """
    draw histogram of data with poisson error bar and f(x,*arg).

    ::

        data = np.random.rand(10000)
        f = gaussian
        draw_compare_hist(f, {'mean':0,'sigma':1}, data, normed=True)

    **Arguments**

        - **f**
        - **arg** argument pass to f. Can be dictionary or list.
        - **data** data array
        - **bins** number of bins. Default 100.
        - **bound** optional boundary of plot in tuple form. If `None` is
          given, the bound is determined from min and max of the data. Default
          `None`
        - **weights** weights array. Default None.
        - **normed** optional normalized data flag. Default False.
        - **use_w2** scaled error down to the original statistics instead of
          weighted statistics.
        - **parts** draw parts of pdf. (Works with AddPdf and Add2PdfNorm).
          Default False.
    """
    ax = plt.gca() if ax is None else ax
    bound = minmax(data) if bound is None else bound
    h, e = np.histogram(data, bins=bins, range=bound, weights=weights)
    err = None
    if weights is not None and use_w2:
        err, _ = np.histogram(data, bins=bins, range=bound,
                              weights=weights * weights)
        err = np.sqrt(err)
    else:
        err = np.sqrt(h)
    return draw_compare(f, arg, e, h, err, ax=ax, grid=grid, normed=normed, parts=parts)



from glob import glob
import os
from setuptools import setup
from setuptools.extension import Extension

import numpy as np

try:
    from Cython.Build import cythonize

    USE_CYTHON = True
except ImportError:
    print('Cython is not available; using pre-generated C files')
    USE_CYTHON = False

ext = '.pyx' if USE_CYTHON else '.c'
extensions = []
for source_file in glob('probfit/*' + ext):
    fname, _ = os.path.splitext(os.path.basename(source_file))
    extensions.append(
        Extension('probfit.{0}'.format(fname),
                  sources=['probfit/{0}{1}'.format(fname, ext)],
                  include_dirs=[np.get_include()])
    )

if USE_CYTHON:
    extensions = cythonize(extensions)


def get_version():
    version = {}
    with open('probfit/version.py') as fp:
        exec(fp.read(), version)
    return version['__version__']


__version__ = get_version()

setup(
    name='probfit',
    version=__version__,
    description='Distribution Fitting/Regression Library',
    long_description=''.join(open('README.rst').readlines()[4:]),
    author='Piti Ongmongkolkul',
    author_email='piti118@gmail.com',
    url='https://github.com/scikit-hep/probfit',
    package_dir={'probfit': 'probfit'},
    packages=['probfit'],
    ext_modules=extensions,
    install_requires=[
        'setuptools',
        'numpy',
        'iminuit',
        'matplotlib'
    ],
    classifiers=[
        "Programming Language :: Python",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Intended Audience :: Science/Research',
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License'
    ],
)



"""
Python 2 / 3 compatibility helpers.
"""
import sys

py_ver = sys.version_info
PY2 = False
PY3 = False
if py_ver[0] == 2:
    PY2 = True
else:  # just in case PY4
    PY3 = True


if PY2:
    range = xrange  # pylint: disable=undefined-variable
else:
    range = range



#cython: embedsignature=True
cimport cython
import numpy as np
cimport numpy as np
from libc.math cimport exp, pow, fabs, log, tgamma, lgamma, sqrt
from matplotlib import pyplot as plt
from . import plotting
from ._libstat cimport compute_nll, compute_chi2_f, compute_bin_chi2_f, \
    csum, compute_bin_lh_f, integrate1d
from .funcutil import FakeFuncCode, merge_func_code
from .nputil import float2double, mid, minmax
from .functor cimport construct_arg
from .util import describe, remove_prefix

np.import_array()

cdef extern from "math.h":
    bint isnan(double x)

cdef class SimultaneousFit:
    """
    Construct simultaneous fit from given cost functions.

    **Arguments**

        - **factors** Optional factor array. If not None, each cost function
          is scaled by `factors[i]` before being summed up. Default None.
        - **prefix** Optional list of prefix. Add prefix to variable name of
          each cost function so that you don't accidentally merge them.
          Default None.
        - **skip_prefix** list of variable names that prefix should not be
          applied to. Useful if you want to prefix them and shared some
          of them.
    """
    cdef readonly list allf
    cdef readonly list allpos
    cdef readonly int numf
    cdef readonly object func_code
    cdef readonly object func_defaults
    cdef np.ndarray factors
    cdef readonly object prefix
    #FIXME: cache each part if called with same parameter
    def __init__(self, *arg, factors=None, prefix=None, skip_prefix=None):
        self.allf = list(arg)
        func_code, allpos = merge_func_code(*arg, prefix=prefix,
                                            skip_prefix=skip_prefix)
        self.numf = len(arg)
        self.func_code = func_code
        self.func_defaults = None
        self.allpos = allpos
        if factors is None:
            factors = np.array([1.] * len(arg))
        self.prefix = prefix
        self.factors = factors

    def __call__(self, *arg):
        cdef double ret = 0.
        cdef np.ndarray[np.int_t] pos
        cdef tuple thisarg
        for i in range(self.numf):
            pos = self.allpos[i]
            thisarg = construct_arg(arg, pos)
            ret += self.factors[i] * self.allf[i](*thisarg)
        return ret

    def args_and_error_for(self, findex, minuit=None, args=None, errors=None):
        """
        convert argument from minuit/dictionary/errors to argument and error
        for cost function with index **findex**
        """
        #dictionary lookup for minuit.values and minuit.errors
        ret_val = None
        ret_err = None
        i = findex
        if minuit is not None:
            keys = minuit.parameters
            p = self.prefix[findex] if self.prefix is not None else None
            #values = dict((remove_prefix(k, p), v) for k,v in minuit.values.items())
            keys = [minuit.parameters[j] for j in self.allpos[i]]
            ret_val = construct_arg(tuple(minuit.args), self.allpos[i])
            errors = dict((remove_prefix(k, p), minuit.errors[k]) for k in keys)
            return ret_val, errors

        if isinstance(args, dict):
            parameters = describe(self)
            pos = self.allpos[i]
            ret_val = [args[parameters[pos[j]]] for j in range(len(pos))]
        else:
            ret_val = construct_arg(args, self.allpos[i])

        if errors is not None:
            ret_err = errors

        return ret_val, ret_err

    def show(self, m=None):
        """
        Same thing as :meth:`draw`. But show the figure immediately.

        .. seealso::
            :meth:`draw` for arguments.

        """
        ret = self.draw(m)
        plt.show()
        return ret

    def draw(self, minuit=None, args=None, errors=None, **kwds):
        """
        Draw each pdf along with data on plotting grid

        return list of return values from each cost function

        """
        return plotting.draw_simultaneous(self, minuit=minuit, args=args,
                                          errors=errors, **kwds)

cdef class UnbinnedLH:
    cdef readonly object f
    cdef readonly object weights
    cdef public object func_code
    cdef readonly np.ndarray data
    cdef readonly int data_len
    cdef readonly badvalue
    cdef readonly tuple last_arg
    cdef readonly bint extended
    cdef readonly tuple extended_bound
    cdef readonly int extended_nint
    def __init__(self, f, data, weights=None, extended=False,
                 extended_bound=None, extended_nint=100, badvalue=-100000):
        """
        Construct -log(unbinned likelihood) from callable *f*
        and data points *data*. Currently can only do 1D fit.

        .. math::
            \\textrm{UnbinnedLH} = \sum_{x \in \\textrm{data}} - \log f(x, arg, \ldots)

        **Arguments**

            - **f** callable object. PDF that describe the data. The parameters
              are parsed using iminuit's ``describe``. The first positional
              arguement is assumed to be independent parameter. For example:

              ::

                    def gauss(x, mu, sigma):#good
                        pass
                    def bad_gauss(mu, sigma, x):#bad
                        pass

            - **data** 1D array of data.
            - **weights** Optional 1D array of weights. Default None(all 1).
            - **badvalue** Optional number. The value that will be used to
              represent log(lh) (notice no minus sign). When the likelihood
              is <= 0. This usually indicate your PDF is faraway from minimum
              or your PDF parameter has gone into unphysical region and return
              negative probability density. This should be a large negative
              number so that iminuit will avoid those points. Default -100000.
            - **extended** Set to True for extended fit. Default False. If Set
              to True the following term is added to resulting negative log
              likelihood

              .. math::
                \\textrm{ext term} = \\int_{x \in \\textrm{extended bound}}f(x, \\textrm{args}, \\ldots) \\textrm{d} x

            - **extended_bound** Bound for calculating extended term.
              Default None(minimum and maximum of data will be used).
            - **extended_nint** number pieces to sum up as
              integral for extended term (using simpson3/8). Default 100.

        .. note::
            There is a notable lack of **sum_w2** for unbinned likelihood. I
            feel like the solutions are quite sketchy. There are multiple ways
            to implement it but they don't really scale correctly. If you
            feel like there is a correct way to implement it feel free to do so
            and write document telling people about the caveat.
        """
        self.f = f
        self.func_code = FakeFuncCode(f, dock=True)
        self.weights = weights
        #only make copy when type mismatch
        self.data = float2double(data)
        self.data_len = len(data)
        self.badvalue = badvalue
        self.extended = extended
        self.extended_bound = extended_bound
        self.extended_nint = extended_nint
        if extended and extended_bound is None:
            self.extended_bound = minmax(data)

    def __call__(self, *arg):
        """
        Compute sum of -log(lh) given positional arguments.
        Position argument will be passed to pdf with independent vairable
        from `data` is given as the frist argument.
        """
        cdef double nll = 0.
        cdef double extended_term = 0.
        self.last_arg = arg
        nll = compute_nll(self.f, self.data, self.weights, arg, self.badvalue)
        if self.extended:
            extended_term = integrate1d(self.f, self.extended_bound,
                                        self.extended_nint, arg)
            nll += extended_term
        return nll

    def draw(self, minuit=None, bins=100, ax=None, bound=None,
             parmloc=(0.05, 0.95), nfbins=200, print_par=True, args=None,
             errors=None, parts=False, show_errbars='normal', no_plot=False):
        """
        Draw comparison between histogram of data and pdf.

        **Arguments**

            - **minuit** Optional but recommended ``iminuit.Minuit`` object.
              If minuit is not ``None``, the pdf will be drawn using minimum
              value from minuit and parameters and error will be shown.
              If minuit is ``None``, then pdf will be drawn using argument from
              the last call to ``__call__``. Default ``None``

            - **bins** number of bins for histogram. Default 100.

            - **ax** matplotlib axes. If not given it will be drawn on current
              axes ``gca()``.

            - **bound** bound for histogram. If ``None`` is given the bound
              will be automatically determined from the data.
              If you given PDF that's normalied to a region but some data is
              not within the bound the picture may look funny.

            - **parmloc** location of parameter print out. This is passed
              directy to legend loc named parameter. Default (0.05,0.95).

            - **nfbins** how many point pdf should be evaluated. Default 200.

            - **print_par** print parameters and error on the plot. Default
              True.

            - **args** Optional. If minuit is not given, parameter value is
              determined from args. This can be dictionary of the form
              `{'a':1.0, 'b':1.0}` or list of values. Default None.

            - **errors** Optional dictionary of errors. If minuit is not given,
              parameter errors are determined from **errors**. Default None.

            - **show_errbars** Show error bars. Default 'normal'

              * 'normal' : error = sqrt( sum of weight )
              * 'sumw2'  : error = sqrt( sum of weight**2 )
              * None : no errorbars (shown as a step histogram)

            - **no_plot** Set this to True if you only want the return value

        **Returns**

        ((data_edges, datay), (errorp,errorm), (total_pdf_x, total_pdf_y), parts)

        """
        return plotting.draw_ulh(self, minuit=minuit, bins=bins, ax=ax,
                                 bound=bound, parmloc=parmloc, nfbins=nfbins, print_par=print_par,
                                 args=args, errors=errors, parts=parts, show_errbars=show_errbars,
                                 no_plot=no_plot)

    def draw_residual(self, minuit=None, bins=100, ax=None, bound=None,
                      parmloc=(0.05, 0.95), print_par=False, args=None, errors=None,
                      errbar_algo='normal', norm=False, **kwargs):
        """
        Draw difference between data and PDF

        **Arguments**

            - **minuit** Optional but recommended ``iminuit.Minuit`` object.
              If minuit is not ``None``, the pdf will be drawn using minimum
              value from minuit and parameters and error will be shown.
              If minuit is ``None``, then pdf will be drawn using argument from
              the last call to ``__call__``. Default ``None``

            - **bins** number of bins for histogram. Default 100.

            - **ax** matplotlib axes. If not given it will be drawn on current
              axes ``gca()``.

            - **bound** bound for histogram. If ``None`` is given the bound
              will be automatically determined from the data.
              If you given PDF that's normalied to a region but some data is
              not within the bound the picture may look funny.

            - **parmloc** location of parameter print out. This is passed
              directy to legend loc named parameter. Default (0.05,0.95).

            - **print_par** print parameters and error on the plot. Default
              True.

            - **args** Optional. If minuit is not given, parameter value is
              determined from args. This can be dictionary of the form
              `{'a':1.0, 'b':1.0}` or list of values. Default None.

            - **errors** Optional dictionary of errors. If minuit is not given,
              parameter errors are determined from **errors**. Default None.

            - **show_errbars** Show error bars. Default True

            - **errbar_algo** How the error bars are calculated
              'normal' : error = sqrt( sum of weight )  [Default]
              'sumw2'  : error = sqrt( sum of weight**2 )

            - **norm** Normalzed by the error bar or not. Default False.

            - **kwargs** Passed to :meth:`probfit.plotting.draw_residual`

        """
        return plotting.draw_residual_ulh(self, minuit=minuit, bins=bins, ax=ax,
                   bound=bound, parmloc=parmloc, print_par=print_par, args=args,
                   errors=errors, errbar_algo=errbar_algo, norm=norm, **kwargs)

    def default_errordef(self):
        return 0.5

    def show(self, *arg, **kwd):
        """
        Same thing as :meth:`draw`. But show the figure immediately.

        .. seealso::
            :meth:`draw` for arguments.

        """
        ret = self.draw(*arg, **kwd)
        plt.show()
        return ret

cdef class BinnedLH:
    cdef readonly object f
    cdef readonly object vf
    cdef readonly object func_code
    cdef readonly np.ndarray h
    cdef readonly np.ndarray w
    cdef readonly np.ndarray w2
    cdef readonly double N
    cdef readonly np.ndarray edges
    cdef readonly np.ndarray midpoints
    cdef readonly np.ndarray binwidth
    cdef readonly int bins
    cdef readonly double mymin
    cdef readonly double mymax
    cdef readonly double badvalue
    cdef readonly tuple last_arg
    cdef readonly int ndof
    cdef readonly bint extended
    cdef readonly bint use_w2
    cdef int nint_subdiv
    def __init__(self, f, data, bins=40, weights=None, weighterrors=None, bound=None,
                 badvalue=1000000, extended=False, use_w2=False, nint_subdiv=1):
        """
        Create a Poisson Binned Likelihood object from given PDF **f** and
        **data** (raw points not histogram). Constant term and expected minimum
        are subtracted off (aka. log likelihood ratio). The exact calculation
        will depend on **extended** and **use_w2** keyword parameters.

        .. math::
            \\textrm{BinnedLH} = -\sum_{i \in bins} s_i \\times  \left(  h_i \\times \log (\\frac{E_i}{h_i}) + (h_i-E_i) \\right)

        where

            - :math:`h_i` is sum of weight of data in ith bin.
            - :math:`b_i` is the width of ith bin.
            - :math:`N` is total number of data. :math:`N = \sum_i h_i`.
            - :math:`E_i` is expected number of occupancy in ith bin from PDF
              calculated using average of pdf value at both sides of the bin
              :math:`l_i, r_i`. The definition for :math:`E_i` depends whether
              extended likelihood is requested.

              If extended likelihood is requested (``extended=True``):

              .. math::
                    E_i = \\frac{f(l_i, arg\ldots )+f(r_i, arg \ldots )}{2} \\times b_i

              If extended likelihood is NOT requested (``extended=False``):

              .. math::
                    E_i = \\frac{f(l_i, arg \ldots )+f(r_i, arg \ldots )}{2} \\times b_i \\times N

              .. note::
                    You are welcome to patch this with a the using real area.
                    So that, it's less sensitive to bin size. Last time I check
                    ROOFIT used **f** evaluated at midpoint.

            - :math:`s_i` is a scaled factor. It's 1 if ``sum_w2=False``.
              It's :math:`s_i = \\frac{h_i}{\sum_{j \in \\textrm{bin }i} w_j^2}`
              if ``sum_w2=True``. The factor will scale the statistics to the
              unweighted data.

            .. note::
                You may wonder why there is :math:`h_i-E_i` added at the end
                for each term of the sum. They sum up to zero anyway.
                The reason is the precision near the minimum. If we taylor
                expand the logarithmic term near :math:`h_i\\approx E_i` then
                the first order term will be :math:`h_i-E_i`. Subtracting this
                term at the end gets us the nice pure parabolic behavior for
                each term at the minimum.

        **Arguments**

            - **f** callable object. PDF that describe the data. The parameters
              are parsed using iminuit's ``describe``. The first positional
              arguement is assumed to be independent parameter. For example:

              ::

                    def gauss(x, mu, sigma):#good
                        pass
                    def bad_gauss(mu, sigma, x):#bad
                        pass

            - **data** 1D array of data. This is raw data not histogrammed
              data.

            - **bins** number of bins data should be histogrammed. Default 40.

            - **weights** Optional 1D array of weights. Default ``None``
              (all 1's).

            - **weighterrors** Optional 1D array of weight errors. Default ``None``
              This is usually used for binned datasets where you want to manipulate
              each bin's error. It doesn't make sense if the data is unbinned data.

            - **bound** tuple(min,max). Histogram bound. If ``None`` is given,
              bound is automatically determined from data. Default None.

            - **badvalue** Optional number. The value that will be used to
              represent log(lh) (notice no minus sign). When the likelihood
              is <= 0. This usually indicate your PDF is faraway from minimum
              or your PDF parameter has gone into unphysical region and return
              negative probability density. This should be a large POSITIVE
              number so that iminuit will avoid those points. Default 100000.

            - **extended** Boolean whether this likelihood should be extended
              likelihood or not. Default False.

            - **use_w2** Scale -log likelihood so that to the original
              unweighted statistics. Default False.

            - **nint_subdiv** controls how BinnedLH do the integral to find
              expect number of event in each bin. The number represent the
              number of subdivisions in each bin to do simpson3/8 rule.
              Default 1.

        """
        self.f = f
        self.func_code = FakeFuncCode(f, dock=True)
        self.use_w2 = use_w2
        self.extended = extended

        if bound is None: bound = minmax(data)

        self.mymin, self.mymax = bound

        h, self.edges = np.histogram(data, bins, range=bound, weights=weights)

        self.h = float2double(h)
        self.N = csum(self.h)

        if weights is not None:
            if weighterrors is None:
                self.w2, _ = np.histogram(data, bins, range=bound,
                                          weights=weights * weights)
            else:
                self.w2, _ = np.histogram(data, bins, range=bound,
                                          weights=weighterrors * weighterrors)
        else:
            self.w2, _ = np.histogram(data, bins, range=bound, weights=None)

        self.w2 = float2double(self.w2)
        self.midpoints = mid(self.edges)
        self.binwidth = np.diff(self.edges)

        self.bins = bins
        self.badvalue = badvalue
        self.ndof = self.bins - (self.func_code.co_argcount - 1)

        self.nint_subdiv = nint_subdiv

    def __call__(self, *arg):
        """
        Calculate sum -log(poisson binned likelihood) given positional
        arguments
        """
        self.last_arg = arg
        ret = compute_bin_lh_f(self.f,
                               self.edges,
                               self.h,  #histogram,
                               self.w2,
                               self.N,  #sum of h
                               arg, self.badvalue,
                               self.extended, self.use_w2,
                               self.nint_subdiv)
        return ret

    def draw(self, minuit=None, ax = None,
             parmloc=(0.05, 0.95), nfbins=200, print_par=True,
             args=None, errors=None, parts=False, no_plot=False):
        """
        Draw comparison between histogram of data and pdf.

        **Arguments**

            - **minuit** Optional but recommended ``iminuit.Minuit`` object.
              If minuit is not ``None``, the pdf will be drawn using minimum
              value from minuit and parameters and error will be shown.
              If minuit is ``None``, then pdf will be drawn using argument from
              the last call to ``__call__``. Default ``None``

            - **ax** matplotlib axes. If not given it will be drawn on current
              axes ``gca()``.

            - **parmloc** location of parameter print out. This is passed
              directy to legend loc named parameter. Default (0.05,0.95).

            - **nfbins** how many point pdf should be evaluated. Default 200.

            - **print_par** print parameters and error on the plot.
              Default True.

            - **no_plot** Set this to True if you only want the return value

        **Returns**

        ((data_edges, data_y), (errorp,errorm), (total_pdf_x, total_pdf_y), parts)

        """
        return plotting.draw_blh(self, minuit=minuit,
                                 ax=ax, parmloc=parmloc, nfbins=nfbins, print_par=print_par,
                                 args=args, errors=errors, parts=parts, no_plot=no_plot)

    def draw_residual(self, minuit=None, ax = None, parmloc=(0.05,0.95),
                      print_par=False, args=None, errors=None, norm=False,
                      **kwargs):
        """
        Draw difference between data and pdf.

        **Arguments**

            - **minuit** Optional but recommended ``iminuit.Minuit`` object.
              If minuit is not ``None``, the pdf will be drawn using minimum
              value from minuit and parameters and error will be shown.
              If minuit is ``None``, then pdf will be drawn using argument from
              the last call to ``__call__``. Default ``None``

            - **ax** matplotlib axes. If not given it will be drawn on current
              axes ``gca()``.

            - **parmloc** location of parameter print out. This is passed
              directy to legend loc named parameter. Default (0.05,0.95).

            - **print_par** print parameters and error on the plot.
              Default True.

            - **norm** If True, draw difference normalized by error
              Default False.

            - **kwargs** Passed to :meth:`probfit.plotting.draw_residual`
        """
        return plotting.draw_residual_blh(self, minuit=minuit,
                                          ax=ax, parmloc=parmloc, print_par=print_par,
                                          args=args, errors=errors, norm=norm,
                                          **kwargs)

    def default_errordef(self):
        return 0.5

    def show(self, *arg, **kwd):
        """
        Same thing as :meth:`draw`. But show the figure immediately.

        .. seealso::
            :meth:`draw` for arguments.

        """
        ret = self.draw(*arg, **kwd)
        plt.show()
        return ret

#fit a line with given function using minimizing chi2
cdef class Chi2Regression:
    cdef readonly object f
    cdef readonly object weights
    cdef readonly object error
    cdef readonly object func_code
    cdef readonly int data_len
    cdef readonly double badvalue
    cdef readonly int ndof
    cdef readonly np.ndarray x
    cdef readonly np.ndarray y
    cdef readonly tuple last_arg

    def __init__(self, f, x, y, error=None, weights=None):
        """
        Create :math:`\chi^2` regression object. This is for fitting funciton
        to data points(x,y) rather than fitting PDF to a distribution.

        .. math::
            \\textrm{Chi2Regression} = \sum_{i} w_i \\times \left( \\frac{f(x_i, arg \ldots) - y_i}{error_i} \\right)^2

        **Arguments**

            - **f** callable object to describe line given by (**x** , **y**).
              The first positional arugment of **f** is assumed to be
              independent variable. Ex:::

                    def gauss(x, mu, sigma):#good
                        pass
                    def bad_gauss(mu, sigma, x):#bad
                        pass

            - **x** 1D array of independent variable
            - **y** 1D array of expected **y**
            - **error** optional 1D error array. If ``None`` is given, it's
              assumed to be all 1's.
            - **weight** 1D array weight for each data point.
        """
        self.f = f
        self.func_code = FakeFuncCode(f, dock=True)
        self.weights = float2double(weights)
        self.error = float2double(error)
        self.x = float2double(x)
        self.y = float2double(y)
        self.data_len = len(x)
        self.ndof = self.data_len - len(describe(self))

    def __call__(self, *arg):
        """
        Compute :math:`\chi^2`
        """
        self.last_arg = arg
        return compute_chi2_f(self.f, self.x, self.y, self.error,
                              self.weights, arg)

    def default_errordef(self):
        return 1.0

    def draw(self, minuit=None, ax=None, parmloc=(0.05, 0.95), print_par=True,
             args=None, errors=None, parts=False, no_plot=False):
        """
        Draw comparison between points (**x**,**y**) and the function **f**.

        **Arguments**

            - **minuit** Optional but recommended ``iminuit.Minuit`` object.
              If minuit is not ``None``, the pdf will be drawn using minimum
              value from minuit and parameters and error will be shown.
              If minuit is ``None``, then pdf will be drawn using argument from
              the last call to ``__call__``. Default ``None``

            - **ax** matplotlib axes. If not given it will be drawn on current
              axes ``gca()``.

            - **parmloc** location of parameter print out. This is passed
              directy to legend loc named parameter. Default (0.05,0.95).

            - **print_par** print parameters and error on the plot.
              Default True.

            - **parts** draw components of PDF. Default False.

            - **no_plot** Set this to true if you only want the return value

        **Returns**

        ((data_x, data_y), (errorp,errorm), (total_pdf_x, total_pdf_y), parts)
        """
        return plotting.draw_x2(self, minuit=minuit, ax=ax, parmloc=parmloc,
                                print_par=print_par, args=args, errors=errors, parts=parts,
                                no_plot=no_plot)

    def show(self, *arg, **kwd):
        """
        Same thing as :meth:`draw`. But show the figure immediately.

        .. seealso::
            :meth:`draw` for arguments.

        """
        ret = self.draw(*arg, **kwd)
        plt.show()
        return ret

    def draw_residual(self, minuit=None, ax=None, args=None, errors=None, grid=True,
                      norm=False):
        plotting.draw_x2_residual(self, minuit=minuit, ax=ax, args=args, errors=errors,
                                  grid=grid, norm=norm)

cdef class BinnedChi2:
    cdef readonly object f
    cdef readonly object vf
    cdef readonly object func_code
    cdef readonly np.ndarray h
    cdef readonly np.ndarray err
    cdef readonly np.ndarray edges
    cdef readonly np.ndarray midpoints
    cdef readonly np.ndarray binwidth
    cdef readonly int bins
    cdef readonly double mymin
    cdef readonly double mymax
    cdef readonly tuple last_arg
    cdef readonly int ndof
    cdef int nint_subdiv
    def __init__(self, f, data, bins=40, weights=None, bound=None,
                 sumw2=False, nint_subdiv=1):
        """
        Create Binned Chi2 Object. It calculates chi^2 assuming poisson
        statistics.

        .. math::
            \\textrm{BinnedChi2} = \sum_{i \in \\textrm{bins}} \left( \\frac{h_i - f(x_i,arg \ldots) \\times b_i \\times N}{err_i} \\right)^2

        Where :math:`err_i` is

            - :math:`\sqrt{\sum_{j \in \\textrm{bin}_i} w_j}` if ``sum_w2=False``.
            - :math:`\sqrt{\sum_{j \in \\textrm{bin}_i} w_j^2}` if ``sum_w2=True``.

        **Arguments**

            - **f** callable object. PDF describing **data**. The first
              positional arugment of **f** is assumed to be independent
              variable. Ex:::

                    def gauss(x, mu, sigma):#good
                        pass
                    def bad_gauss(mu, sigma, x):#bad
                        pass

            - **data** 1D array data (raw not histogram)

            - **bins** Optional number of bins to histogram data. Default 40.

            - **weights** 1D array weights.

            - **bound** tuple(min,max) bound of histogram. If ``None`` is given
              it's automatically determined from the data.

            - **sumw2** scale the error using
              :math:`\sqrt{\sum_{j \in \\textrm{bin}_i} w_j^2}`.

            - **nint_subdiv** controls how BinnedChi2 do the integral to find
              expect number of event in each bin. The number represent the
              number of subdivisions in each bin to do simpson3/8.
              Default 1.
        """
        self.f = f
        self.func_code = FakeFuncCode(f, dock=True)
        if bound is None:
            bound = minmax(data)
        self.mymin, self.mymax = bound

        h, self.edges = np.histogram(data, bins, range=bound, weights=weights)

        self.h = float2double(h)
        self.midpoints = mid(self.edges)
        self.binwidth = np.diff(self.edges)

        #sumw2 if requested
        if weights is not None and sumw2:
            w2 = weights * weights
            sw2, _ = np.histogram(data, bins, range=bound, weights=w2)
            self.err = np.sqrt(sw2)
        else:
            self.err = np.sqrt(self.h)

        #check if error is too small
        if np.any(self.err < 1e-5):
            raise ValueError('some bins are too small to do a chi2 fit. change your range')

        self.bins = bins
        self.ndof = self.bins - len(describe(self))  # fix this taking care of fixed parameter
        self.nint_subdiv = nint_subdiv

    #lazy mid point implementation
    def __call__(self, *arg):
        """
        Calculate :math:`\chi^2` given positional arguments
        """
        self.last_arg = arg
        return compute_bin_chi2_f(self.f, self.edges, self.h, self.err,
                                  None, arg, self.nint_subdiv)

    def draw(self, minuit=None, ax = None, parmloc=(0.05, 0.95), nfbins=200,
             print_par=True, args=None, errors=None, parts=False, no_plot=False):
        """
        Draw comparison histogram of data and the function **f**.

        **Arguments**

            - **minuit** Optional but recommended ``iminuit.Minuit`` object.
              If minuit is not ``None``, the pdf will be drawn using minimum
              value from minuit and parameters and error will be shown.
              If minuit is ``None``, then pdf will be drawn using argument from
              the last call to ``__call__``. Default ``None``

            - **ax** matplotlib axes. If not given it will be drawn on current
              axes ``gca()``.

            - **parmloc** location of parameter print out. This is passed
              directy to legend loc named parameter. Default (0.05,0.95).

            - **nfbins** number of points to calculate f

            - **print_par** print parameters and error on the plot.
              Default True.

            - **no_plot** Set this to true if you only want the return value

        **Returns**

        ((data_edges, data_y), (errorp,errorm), (total_pdf_x, total_pdf_y), parts)
        """
        return plotting.draw_bx2(self, minuit=minuit, ax=ax,
                                 parmloc=parmloc, nfbins=nfbins, print_par=print_par,
                                 args=args, errors=errors, parts=parts, no_plot=no_plot)

    def default_errordef(self):
        return 1.0

    def show(self, *arg, **kwd):
        """
        Same thing as :meth:`draw`. But show the figure immediately.

        .. seealso::
            :meth:`draw` for arguments.

        """
        ret = self.draw(*arg, **kwd)
        plt.show()
        return ret



import itertools as itt
import collections
import numpy as np
from matplotlib import pyplot as plt
from iminuit import Minuit
from .py23_compat import range
from .costfunc import UnbinnedLH, BinnedChi2, BinnedLH
from ._libstat import _vector_apply
from .nputil import minmax


def fit_uml(f, data, quiet=False, print_level=0, *arg, **kwd):
    """
    perform unbinned likelihood fit
    :param f: pdf
    :param data: data
    :param quiet: if not quite draw latest fit on fail fit
    :param printlevel: minuit printlevel
    :return:
    """
    uml = UnbinnedLH(f, data)
    minuit = Minuit(uml, print_level=print_level, **kwd)
    minuit.set_strategy(2)
    minuit.migrad()
    if not minuit.migrad_ok() or not minuit.matrix_accurate():
        if not quiet:
            plt.figure()
            uml.show()
            print(minuit.values)
    return (uml, minuit)


def fit_binx2(f, data, bins=30, bound=None, print_level=0, quiet=False, *arg, **kwd):
    """
    perform chi^2 fit
    :param f:
    :param data:
    :param bins:
    :param range:
    :param printlevel:
    :param quiet:
    :param arg:
    :param kwd:
    :return:
    """
    uml = BinnedChi2(f, data, bins=bins, bound=bound)
    minuit = Minuit(uml, print_level=print_level, **kwd)
    minuit.set_strategy(2)
    minuit.migrad()
    if not minuit.migrad_ok() or not minuit.matrix_accurate():
        if not quiet:
            plt.figure()
            uml.show()
            print(minuit.values)

    return (uml, minuit)


def fit_binlh(f, data, bins=30,
              bound=None, quiet=False, weights=None, use_w2=False,
              print_level=0, pedantic=True, extended=False,
              *arg, **kwd):
    """
    perform bin likelihood fit
    :param f:
    :param data:
    :param bins:
    :param range:
    :param quiet:
    :param weights:
    :param use_w2:
    :param printlevel:
    :param pedantic:
    :param extended:
    :param arg:
    :param kwd:
    :return:
    """
    uml = BinnedLH(f, data, bins=bins, bound=bound,
                   weights=weights, use_w2=use_w2, extended=extended)
    minuit = Minuit(uml, print_level=print_level, pedantic=pedantic, **kwd)
    minuit.set_strategy(2)
    minuit.migrad()
    if not minuit.migrad_ok() or not minuit.matrix_accurate():
        if not quiet:
            plt.figure()
            uml.show()
            print(minuit.values)
    return (uml, minuit)


def tuplize(x):
    """
    :param x:
    :return:
    """
    if isinstance(x, collections.Iterable):
        return x
    else:
        return tuple([x])


def pprint_arg(vnames, value):
    """
    pretty print argument
    :param vnames:
    :param value:
    :return:
    """
    ret = ''
    for name, v in zip(vnames, value):
        ret += '%s=%s;' % (name, str(v))
    return ret;


def try_uml(f, data, bins=40, fbins=1000, *arg, **kwd):
    fom = UnbinnedLH(f, data)
    narg = f.func_code.co_argcount
    vnames = f.func_code.co_varnames[1:narg]
    my_arg = [tuplize(kwd[name]) for name in vnames]
    h, e, _ = plt.hist(data, bins=bins, normed=True, histtype='step')
    vx = np.linspace(e[0], e[-1], fbins)
    first = True
    minfom = 0
    minarg = None
    for thisarg in itt.product(*my_arg):
        vy = _vector_apply(f, vx, thisarg)
        plt.plot(vx, vy, '-', label=pprint_arg(vnames, thisarg))
        thisfom = fom(*thisarg)
        if first or thisfom < minfom:
            minfom = thisfom
            minarg = thisarg
            first = False
    leg = plt.legend(fancybox=True)
    leg.get_frame().set_alpha(0.5)
    ret = dict((k, v) for k, v in zip(vnames, minarg))
    return ret


def try_binlh(f, data, weights=None, bins=40, fbins=1000, show='both', extended=False,
              bound=None, *arg, **kwd):
    if bound is None: bound = minmax(data)
    fom = BinnedLH(f, data, extended=extended, bound=bound)
    narg = f.func_code.co_argcount
    vnames = f.func_code.co_varnames[1:narg]
    my_arg = [tuplize(kwd[name]) for name in vnames]
    h, e = None, None
    if show == 'both':
        h, e, _ = plt.hist(data, bins=bins, range=bound, histtype='step', weights=weights, normed=not extended)
    else:
        h, e = np.histogram(data, bins=bins, range=bound, weights=weights, normed=not extended)
    bw = e[1] - e[0]

    vx = np.linspace(e[0], e[-1], fbins)
    first = True
    minfom = 0
    minarg = None
    for thisarg in itt.product(*my_arg):
        vy = _vector_apply(f, vx, thisarg)
        if extended: vy *= bw
        plt.plot(vx, vy, '-', label=pprint_arg(vnames, thisarg))
        thisfom = fom(*thisarg)
        if first or thisfom < minfom:
            minfom = thisfom
            minarg = thisarg
            first = False
    leg = plt.legend(fancybox=True)
    leg.get_frame().set_alpha(0.5)
    ret = dict((k, v) for k, v in zip(vnames, minarg))
    return ret


def try_chi2(f, data, weights=None, bins=40, fbins=1000, show='both', *arg, **kwd):
    fom = BinnedChi2(f, data)
    narg = f.func_code.co_argcount
    vnames = f.func_code.co_varnames[1:narg]
    my_arg = [tuplize(kwd[name]) for name in vnames]
    h, e = None, None
    if show == 'both':
        h, e, _ = plt.hist(data, bins=bins, histtype='step', weights=weights)
    else:
        h, e = np.histogram(data, bins=bins, weights=weights)
    bw = e[1] - e[0]
    vx = np.linspace(e[0], e[-1], fbins)
    first = True
    minfom = 0
    minarg = None
    for thisarg in itt.product(*my_arg):
        vy = _vector_apply(f, vx, thisarg) * bw
        plt.plot(vx, vy, '-', label=pprint_arg(vnames, thisarg))
        thisfom = fom(*thisarg)
        if first or thisfom < minfom:
            minfom = thisfom
            minarg = thisarg
            first = False
    leg = plt.legend(fancybox=True)
    leg.get_frame().set_alpha(0.5)
    ret = dict((k, v) for k, v in zip(vnames, minarg))
    return ret

# def randfr(r):
#     """
#     generate a uniform random number with in range r
#     :param r: tuple range
#     :return: float
#     """
#     b = r[1]
#     a = r[0]
#     return np.random.ranf() * (b - a) + a

# def guess_initial(alg, f, data, ntry=100, guessrange=(-100, 100), draw=False, *arg, **kwd):
#     """
#     This is very bad at the moment don't use it
#     """
#     fom = alg(f, data, *arg, **kwd)
#     first = True
#     minfom = 0
#     minparam = ()
#     narg = fom.func_code.co_argcount
#     vnames = fom.func_code.co_varnames[:narg]
#     ranges = {}
#     for vname in vnames:
#         if 'limit_' + vname in kwd:
#             ranges[vname] = kwd['limit_' + vname]
#         else:
#             ranges[vname] = guessrange

#     for i in range(ntry):
#         arg = []
#         for vname in vnames:
#             arg.append(randfr(ranges[vname]))
#         try:
#             thisfom = fom(*arg)
#             if first or thisfom < minfom:
#                 first = False
#                 minfom = thisfom
#                 minparam = arg
#         except ZeroDivisionError:
#             pass

#     print(minparam, minfom)
#     ret = {}
#     for vname, bestguess in zip(vnames, minparam):
#         ret[vname] = bestguess
#     if draw:
#         fom(*minparam)
#         fom.draw()
#     return ret
